__all__ = '__title__ __author__ __version__ __description__ __license__ __copyright__'.split()

__title__ = 'vdsql'
__author__ = 'Saul Pwanson <code@saul.pw>'
__version__ = '0.3dev'
__description__ = 'VisiData for database queries'
__license__ = 'Apache License, Version 2.0'
__copyright__ = 'Copyright 2022 ' + __author__
